#ifndef Array_h
#define Array_h

// if RANGECHECK is defined, RANGE_CHECK_HANDLER should be a function that
// handles out-of-bounds problems; NULL_POINTER_HANDLER should handle
// null pointer dereferences.  If you are using a newer compiler, these
// functions could throw exceptions.
#ifdef RANGECHECK
void CC RANGE_CHECK_HANDLER( int size, int lo, int hi, int index );
void CC NULL_POINTER_HANDLER( int size, int pointertype );
#endif

template <class T,int lo,int hi>
class Array
{
  protected:
	T data[hi-lo+1];
  public:
	inline T& operator [] ( int index )
	{
#ifdef RANGECHECK
		if( (unsigned)(index-lo) > (unsigned)(hi-lo) )
			RANGE_CHECK_HANDLER( sizeof(T), lo, hi, index );
#endif
		return data[index-lo];
	}

	inline T operator [] ( int index ) const
	{
#ifdef RANGECHECK
		if( (unsigned)(index-lo) > (unsigned)(hi-lo) )
			RANGE_CHECK_HANDLER( sizeof(T), lo, hi, index );
#endif
		return data[index-lo];
	}
};

template <class T,int lo,int hi>
class AllocArray
{
  protected:
	T* data;
  public:
	AllocArray()
	{
		data = new T[hi-lo+1];
#ifdef RANGECHECK
		if( !data )
			NULL_POINTER_HANDLER(sizeof(T),2);
#endif
	}
	~AllocArray()
	{
		delete [] data;
	}
	inline T& operator [] ( int index )
	{
#ifdef RANGECHECK
		if( (unsigned)(index-lo) > (unsigned)(hi-lo) )
			RANGE_CHECK_HANDLER( sizeof(T), lo, hi, index );
#endif
		return data[index-lo];
	}

	inline T operator [] ( int index ) const
	{
#ifdef RANGECHECK
		if( (unsigned)(index-lo) > (unsigned)(hi-lo) )
			RANGE_CHECK_HANDLER( sizeof(T), lo, hi, index );
#endif
		return data[index-lo];
	}
};

// safer pointer
template<class T>
struct Pointer
{
	T* x;
	Pointer( T* y = 0 ): x(y) {}
	Pointer( const Pointer<T>& y ): x(y.x) {}
	T* operator -> () const
	{
#ifdef RANGECHECK
		if( !x )
			NULL_POINTER_HANDLER(sizeof(T),1);
#endif
		return x;
	}
	T& operator * () const
	{
#ifdef RANGECHECK
		if( !x )
			NULL_POINTER_HANDLER(sizeof(T),0);
#endif
		return *x;
	}
	int operator ! () const { return !x; }
	operator T* () const 	{ return x; }
};

// safer pointer
template<class T>
struct SimplePointer
{
	T* x;
	SimplePointer( T* y = 0 ): x(y) {}
	SimplePointer( const SimplePointer<T>& y ): x(y.x) {}
	T& operator * () const
	{
#ifdef RANGECHECK
		if( !x )
			NULL_POINTER_HANDLER(sizeof(T),0);
#endif
		return *x;
	}
	int operator ! () const { return !x; }
	operator T* () const 	{ return x; }
};

#endif
