//////////////////////////////////////////////////////////////////////
// INOUT                                                            //
//////////////////////////////////////////////////////////////////////
// This unit is the heart of SRCOMM -- it provides something        //
// similar to C++ Streams, but designed for door games.  Added      //
// features include color handling, a line buffer that remembers    //
// the current line (so that you can interrupt the user for Chat    //
// and restore the line afterwards), default & maximum values for   //
// numeric input, and support for filters, which let the user       //
// define his own output conversion (so that you can do things like //
// color codes and "flavors" as SRE does).                          //
//////////////////////////////////////////////////////////////////////

#include <ctype.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>

#include <cstring.h>
#include <unit/StrUtl.h>

#include <srcomm3/inout.h>

char standardout::CommaCharacter = ',';

standardin::standardin()
{
    fieldwidth = 0;
    num_default = -1;
    num_max = -1;
}

char CC standardin::out( char c )
{
	return c;
}

void CC standardin::outs( const char * s )
{
	while( *s )
	{
		out( *s );
		s++;
	}
}

void CC standardin::getstr( string& s )
{
    s = "";

	begininput();

    for(;;)
    {
		int ch = (unsigned char)get();
		if( ch == '\t' )
			ch = ' ';

        if( ch == '\b' && length(s) > 0 )
        {
            s.remove( length(s)-1 );
			outs( "\b \b" );
        }
		else if( ch >= ' ' && ch <= 254 && ch != 127 && ( length(s) < width() || width() == 0 ) )
        {
			s += char(ch);
			out( ch );
		}
		if( ch == '\n' || ch == '\r' || ch == '\0' )
		{
			out( '\n' );
			endinput();
			return;
		}
	}
}

void CC standardin::begininput()
{
}

void CC standardin::endinput()
{
}

standardin& CC standardin::setw( int w )
{
	fieldwidth = w;
    return *this;
}

int CC standardin::width()
{
    return fieldwidth;
}

standardin& CC standardin::setdefault( long def )
{
	num_default = def;
	return *this;
}

standardin& CC standardin::setmax( long max )
{
	num_max = max;
	return *this;
}

// standardin class

standardin& CC standardin::operator >> ( char& c )
{
	 c = get();
	 return *this;
}

standardin& CC standardin::operator >> ( string& s )
{
    getstr( s );
	return *this;
}

long CC standardin::readnum( long NUM_MAX )
{
    char ch, i = 0;
    long l = -1;
	int postfix = 0;
    char validchar = 0;

    if( NUM_MAX > num_max && num_max >= 0 ) NUM_MAX = num_max;

	begininput();

	do
	{
        if( postfix == 0 )
            ch = toupper( get() );
        else
        {
			postfix--;
			ch = '0';
		}

		if( ch == '\b' && i > 0 )
		{
			if( l > 9 )
				l /= 10;
			else
				l = -1;
			outs( "\b \b" );
			i--;
            validchar = 1;
		}
		else if( ch >= '0' && ch <= '9' )
        {
            if( l == -1 )
            {
                long nl = ( ch - '0' );
				if( nl > NUM_MAX ) nl = -1;
				if( nl >= 0 )
				{
					i++;
					out( ch );
					l = nl;
                }
			}
			else
			{
                if( l == 0 )
                {
                    out( '\b' );
					i--;
				}

                long nl = l * 10 + ( ch - '0' );
				if( nl > NUM_MAX ) nl = -1;
				else if( nl / 10 != l ) nl = -1;

				if( nl >= 0 )
				{
					i++;
					out( ch );
					l = nl;
				}
                else if( l == 0 )
                    l = -1;
			}
            validchar = 1;
		}
        else if( ch == '>' && num_max >= 0 )
        {
            while( i > 0 )
			{
				outs( "\b \b" );
                i--;
            }
			// count # digits in max
            l = num_max;
            for( i = 0; l > 0; i++ )
				l /= 10;
            for( int j = i-1; j >= 0; j-- )
            {
                l = num_max;
                for( int k = 0; k < j; k++ )
                    l /= 10;
				out( '0' + l%10 );
			}
            l = num_max;
            validchar = 1;
        }
        else if( ch == 'K' || ch == 'M' )
        {
			if( validchar )
            {
                if( ch == 'K' ) postfix = 3;
                if( ch == 'M' ) postfix = 6;
            }
        }
        else
            validchar = 0;

    } while( ch != '\n' && ch != '\r' && ch != '\0' );

    if( i == 0 )  l = -1;

	if( l == -1 )
    {
        string s;

        if( num_default != -1 )
            s = str(num_default);

		outs( s.c_str() );

		l = num_default;
	}

	num_default = -1;
	num_max = -1;

    endinput();
    out( '\n' );

    return l;
}

standardin& CC standardin::operator >> ( long& l )
{
	l = readnum( LONG_MAX );
	return *this;
}

standardin& CC standardin::operator >> ( int& i )
{
	long l;

    l = readnum( INT_MAX );
	if(l==-1)	i = -1;
	else		i = (int)l;

	return *this;
}

standardin& CC standardin::operator >> ( unsigned& i )
{
	i = (unsigned)( readnum(UINT_MAX) & 0xffff );
	return *this;
}

standardin& CC standardin::operator >> ( unsigned char& i )
{
	i = readnum(UCHAR_MAX) & 0xff;
	return *this;
}

// LineBuffer class
LineBuffer::LineBuffer( int w )
{
	if( w > 0 )
	{
		buf = new char[w+2];
		fg = new int[w+2];
		bg = new int[w+2];
	}
	else
	{
		buf = 0;
		fg = 0;
		bg = 0;
	}

	bufsize = linewidth = w;

    if( buf )     buf[0] = 0;
	bufpos = 0;
}

LineBuffer::LineBuffer( const LineBuffer& b )
{
	copy( b );
}

LineBuffer::~LineBuffer()
{
    kill();
}

LineBuffer& LineBuffer::operator = ( const LineBuffer& b )
{
	kill();
	copy( b );
	return *this;
}

void LineBuffer::copy( const LineBuffer& b )
{
	bufsize = linewidth = b.linewidth;
	if( linewidth > 0 )
	{
		buf = new char[linewidth+2];
		fg = new int[linewidth+2];
        bg = new int[linewidth+2];
    }
    else
    {
		buf = 0;
		fg = 0;
        bg = 0;
    }

	if( buf )
	{
		strcpy( buf, b.buf );
        bufpos = b.bufpos;
    }
    else
        bufpos = 0;

	if( fg )
		for( int i = 0; i < bufsize+2; i++ )
			fg[i] = b.fg[i];
	if( bg )
		for( int i = 0; i < bufsize+2; i++ )
			bg[i] = b.bg[i];
}

void LineBuffer::kill()
{
	delete [] buf;
	delete [] bg;
	delete [] fg;
}

void CC LineBuffer::insert( char c, int f, int b )
{
    if( buf && bufpos <= linewidth )
    {
        fg[bufpos] = f;
		bg[bufpos] = b;
        buf[bufpos++] = c;
        buf[bufpos] = '\0';
    }
}

void CC LineBuffer::backspace()
{
	if( buf && bufpos > 0 )
		buf[--bufpos] = '\0';
}

void CC LineBuffer::narrow( int w )
{
	if( w <= bufsize )
		linewidth = w;
	else
		linewidth = bufsize;
}

void LineBuffer::clear()
{
    bufpos = 0;
	if( buf ) buf[0] = '\0';
}

char LineBuffer::operator [] ( int index )
{
    if( buf && index >= 0 && index < linewidth )
        return buf[index];
    return 0;
}

int LineBuffer::fore( int index )
{
    if( fg && index >= 0 && index < linewidth )
		return fg[index];
    return 0;
}

int LineBuffer::back( int index )
{
    if( bg && index >= 0 && index < linewidth )
        return bg[index];
    return 0;
}

standardout& operator << ( standardout& out, const LineBuffer& b )
{
	if( !b.buf ) return out;
    for( int i = 0; i < b.bufpos; i++ )
    {
        if( b.fg && b.bg )
            out << attr( b.fg[i], b.bg[i] );
        if( b.buf )
            out << b.buf[i];
    }

    return out;
}

// standardout class

standardout::standardout( int newwidth ): buffer( newwidth )
{
	changeColorFunc = 0;
    translate = StandardOutputTranslator;

    current_fore = 0;
	current_back = 0;
	new_fore = 7;
	new_back = 0;

    rawmode = 0;
}

standardout::~standardout()
{
}

standardout& CC standardout::operator << ( char c )
{
    put( 1, &c );
    return *this;
}

void StandardOutputTranslator( standardout& out, const string& s )
{
	out.put( length(s), s.c_str() );
}

standardout& CC standardout::operator << ( const string& s0 )
{
	string s = s0;
    if( translate ) translate( *this, s );
	return *this;
}

standardout& CC standardout::operator << ( const char * s0 )
{
	string s = s0;
    if( translate ) translate( *this, s );
	return *this;
}

void standardout::writenum( long l )
{
    string s = strcomma( l, CommaCharacter != 0 );
    for( int i = 0; i < length(s); i++ )
		if( s[i] == ',' ) s(i) = CommaCharacter;
    if( translate ) translate( *this, s );
}

ColorTranslator standardout::setColorTranslator( ColorTranslator t )
{
	ColorTranslator old = changeColorFunc;
	changeColorFunc = t;
    return old;
}

OutputTranslator standardout::setOutputTranslator( OutputTranslator t )
{
    OutputTranslator old = translate;
	if( !t ) t = StandardOutputTranslator;
    translate = t;
    return old;
}

void standardout::changecolor()
{
    if( changeColorFunc )
        changeColorFunc( *this );
    current_fore = new_fore;
    current_back = new_back;
}

void AnsiTranslator( standardout& out )
{
	const char *c = "04261537";

    int new_back = out.GetBackground();
    int new_fore = out.GetForeground();
	int current_back = out.GetActualBackground();
    int current_fore = out.GetActualForeground();

	char s[15];
	int i = 2;
	s[0] = 27;
	s[1] = '[';

    if( new_back == -1 )
        new_back = current_back;

    if( ( current_fore >= 8 && new_fore < 8 ) || ( new_fore == 7 && new_back == 0 ) )
	{
		current_fore = 7;
        current_back = 0;
		s[i++] = '0';
		s[i++] = ';';
	}

    if( current_fore < 8 && new_fore >= 8 )
	{
        current_fore |= 8;
		s[i++] = '1';
		s[i++] = ';';
	}

	if( current_fore != new_fore )
	{
        current_fore = new_fore;
		s[i++] = '3';
        s[i++] = c[ new_fore & 7 ];
		s[i++] = ';';
	}

    if( current_back != new_back )
	{
        current_back = new_back;
		s[i++] = '4';
        s[i++] = c[ new_back & 7 ];
		s[i++] = ';';
	}

	s[i] = '\0';
	i--;
	s[i] = 'm';
	if( i > 1 )
        out._put( i+1, s );
}

void CC standardout::put( int len, const char c[] )
{
	if( current_fore != new_fore || current_back != new_back )
        changecolor();

    if( rawmode )
	{
        const char* s = c;
        for( int i = 0, l = 0; i < len; i++, l++ )
        {
            if( c[i] == '\r' )   // filter out the returns
            {
                _put( l, s );    // not including c[i]
                l = -1;           // don't put in the \r
                s = c+i+1;
            }
            else if( c[i] == '\n' ) // add a return
			{
                _put( l, s );   // not including c[i]
                _put( 2, "\r\n" );
                l = -1;
                s = c+i+1;
            }
        }
        _put( l, s );
        return;
    }

    const char* s = c;
    int w = buffer.width();
    for( int i = 0, l = 0; i < len; i++, l++ )
    {
        if( c[i] == '\r' )   // filter out the returns
        {
            _put( l, s );    // not including c[i]
            l = -1;           // don't put in the \r
            s = c+i+1;
        }
        else if( c[i] == '\n' ) // add a return
        {
			_put( l, s );   // not including c[i]
            _put( 2, "\r\n" );
            l = -1;
            s = c+i+1;
            buffer.clear();
        }
        else if( c[i] == '\b' ) // backspace
        {
            buffer.backspace();
        }
        else
        {
			if( !iscntrl( (unsigned char)c[i] ) )
				buffer.insert( c[i], current_fore, current_back );

            int p = buffer.pos();
            if( w > 0 && p >= w )
            {
                int w34 = w * 3 / 4;

                while( !isspace( (unsigned char)buffer[p] ) && p > w34 )
                    p--;

				_put( l+1, s );
				l = -1;
                s = c+i+1;

                if( p <= w34 )
                    p = buffer.pos();
                else
					p++;

				{
					{
						int f = new_fore, b = new_back;
						SetAttr( f, 0 );
						changecolor();
						for( int j = p; j < buffer.pos(); j++ )
							_put( 3, "\b \b" );
							_put( 2, "\r\n" );
						SetAttr( f, b );
						changecolor();
					}

					LineBuffer b = buffer;
					buffer.clear();

                    buffer.insert( ' ', current_fore, current_back );
                    _put( 1, " " );
                    for( ; ( b[p]==' ' || b[p]=='\t' ) && p < b.pos(); p++ )
                        /* skip leading spaces */ ;
					for( ; p < b.pos(); p++ )
					{
						SetAttr( b.fore(p), b.back(p) );

						char a = b[p];
						if( !iscntrl( (unsigned char)a ) )
							buffer.insert( a, current_fore, current_back );
						_put( 1, &a );
					}
				}
			}
		}
	}
	_put( l, s );
}

void CC standardout::_put( int, const char [] )
{
    return;
}

void CC standardout::SetAttr( int f, int b )
{
    new_fore = f;
    if( b != -1 )
        new_back = b;
    else
        new_back = current_back;
}

// Color object system
CC attr::attr( int f, int b )
	:fore(f), back(b)
{
}

standardout& CC operator << ( standardout& out, attr& a )
{
	out.SetAttr( a.fore, a.back );
	return out;
}

